from typing import Sequence, Union

from redbot.core.utils.chat_formatting import humanize_list, humanize_number, inline


def _hum(num: Union[int, float], unit: str, ndigits: int) -> str:
    """Round a number, then humanize."""
    return humanize_number(round(num, ndigits)) + f" {unit}"


def humanize_bytes(bytes: Union[int, float], ndigits: int = 0) -> str:
    """Humanize a number of bytes, rounding to ndigits. Only supports up to GB.

    This assumes 1GB = 1000MB, 1MB = 1000KB, 1KB = 1000B"""
    if bytes > 10000000000:  # 10GB
        gb = bytes / 1000000000
        return _hum(gb, "GB", ndigits)
    if bytes > 10000000:  # 10MB
        mb = bytes / 1000000
        return _hum(mb, "MB", ndigits)
    if bytes > 10000:  # 10KB
        kb = bytes / 1000
        return _hum(kb, "KB", ndigits)
    return _hum(bytes, "B", 0)  # no point in rounding


# maybe think about adding to core
def inline_hum_list(items: Sequence[str], *, style: str = "standard") -> str:
    """Similar to core's humanize_list, but all items are in inline code blocks. **Can** be used
    outside my cogs.

    Strips leading and trailing whitespace.

    Does not support locale.

    Does support style (see core's docs for available styles)

    Parameters
    ----------
    items : Sequence[str]
        The items to humanize
    style : str, optional
        The style. See core's docs, by default "standard"

    Returns
    -------
    str
        Humanized inline list.
    """
    inline_list = [inline(i.strip()) for i in items]
    return humanize_list(inline_list, style=style)



from typing import Dict, List, Tuple, Union

import aiohttp
import tabulate
from packaging import version  # required by setuptools which is required by red
from redbot.core import commands
from redbot.core.utils.chat_formatting import box

from vexcogutils.loop import VexLoop

from .consts import CHECK, CROSS, DOCS_BASE
from .version import __version__


def format_help(self: commands.Cog, ctx: commands.Context) -> str:
    """Wrapper for format_help_for_context. **Not** currently for use outside my cogs.

    Thanks Sinbad.

    Parameters
    ----------
    self : commands.Cog
        The Cog class
    context : commands.Context
        Context

    Returns
    -------
    str
        Formatted help
    """
    docs = DOCS_BASE.format(self.qualified_name.lower())
    pre_processed = super(type(self), self).format_help_for_context(ctx)

    return (
        f"{pre_processed}\n\nAuthor: **`{self.__author__}`**\nCog Version: "
        f"**`{self.__version__}`**\n{docs}"
    )
    # adding docs link here so doesn't show up in auto generated docs


async def format_info(
    qualified_name: str,
    cog_version: str,
    extras: Dict[str, Union[str, bool]] = {},
    loops: List[VexLoop] = [],
) -> str:
    """Generate simple info text about the cog. **Not** currently for use outside my cogs.

    Parameters
    ----------
    qualified_name : str
        The name you want to show, eg "BetterUptime"
    cog_version : str
        The version of the cog
    extras : Dict[str, Union[str, bool]], optional
        Dict which is foramtted as key: value\\n. Bools as a value will be replaced with
        check/cross emojis, by default {}
    loops : List[VexLoop], optional
        List of VexLoops you want to show, by default []

    Returns
    -------
    str
        Simple info text.
    """
    try:
        latest_cog, latest_utils = await _get_latest_ver(qualified_name.lower())
        cog_updated = CHECK if version.parse(cog_version) >= version.parse(latest_cog) else CROSS
        utils_updated = (
            CHECK if version.parse(__version__) >= version.parse(latest_utils) else CROSS
        )
    except Exception:  # anything and everything
        cog_updated = "Unknown"
        utils_updated = "Unknown"
    start = f"{qualified_name} by Vexed.\n<https://github.com/Vexed01/Vex-Cogs>\n\n"
    data = [
        ["Cog Version", cog_version],
        ["Cog up to date", cog_updated],
        ["Utils Version", __version__],
        ["Utils up to data", utils_updated],
    ]

    if loops:
        data.append([])
        for loop in loops:
            data.append([loop.friendly_name, CHECK if loop.integrity else CROSS])

    if extras:
        data.append([])
        for key, value in extras.items():
            if isinstance(value, bool):
                value = CHECK if value else CROSS
            data.append([key, value])

    boxed = box(tabulate.tabulate(data))

    return f"{start}\n{boxed}"


async def _get_latest_ver(cog_name: str) -> Tuple[str, str]:
    async with aiohttp.ClientSession() as session:
        resp = await session.get(
            "https://vexed01.github.io/Vex-Cogs/api/v1/versions.json", timeout=3  # impatient :aha:
        )
        as_dict: dict = await resp.json()
        session.close()

    latest_cog = as_dict.get("cogs", {}).get(cog_name, "Unknown")
    latest_utils = as_dict.get("utils", "Unknown")
    return latest_cog, latest_utils



DOCS_BASE = (
    "This cog has docs! Check them out at\n"
    "https://vex-cogs.rtfd.io/en/latest/cogs/{}.html?utm_source=cog&utm_medium=docstring"
    "&utm_campaign=main_help"
)

CHECK = "\N{HEAVY CHECK MARK}\N{VARIATION SELECTOR-16}"
CROSS = "\N{CROSS MARK}"



from pathlib import Path

from setuptools import setup

exc: dict = {}
with open(Path(__file__).parent / "vexcogutils" / "version.py") as fp:
    exec(fp.read(), exc)
version = exc["__version__"]


setup(
    name="vex-cog-utils",
    version=version,
    url="https://github.com/Vexed01/vex-cog-utils",
    author="Vexed01",
    author_email="gh.vexed@gmail.com",
    description="Utility functions for the Vex-Cogs repo.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=["vexcogutils"],
    package_data={"vexcogutils": ["py.typed"]},
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8.1",
    install_requires=["tabulate"],
)



import asyncio
import datetime
import traceback
from typing import List, Optional

import discord
import tabulate
from redbot.core.utils.chat_formatting import box, pagify

from vexcogutils.consts import CHECK, CROSS


class VexLoop:
    """
    A class with some utilities for logging the state of a loop.

    Note iter_count increases at the start of an iteration.

    This does not log anything itself.
    """

    def __init__(self, friendly_name: str, expected_interval: float) -> None:
        self.friendly_name = friendly_name
        self.expected_interval = datetime.timedelta(seconds=expected_interval)

        self.iter_count: int = 0
        self.currently_running: bool = False  # whether the loop is running or sleeping
        self.last_exc: str = "No exception has occurred yet."
        self.last_exc_raw: Optional[BaseException] = None

        self.last_iter: Optional[datetime.datetime] = None
        self.next_iter: Optional[datetime.datetime] = None

    def __repr__(self) -> str:
        return (
            f"<friendly_name={self.friendly_name} iter_count={self.iter_count} "
            f"currently_running={self.currently_running} last_iter={self.last_iter} "
            f"next_iter={self.next_iter} integrity={self.integrity}>"
        )

    @property
    def integrity(self) -> bool:
        """
        If the loop is running on time (whether or not next expected iteration is in the future)
        """
        if self.next_iter is None:  # not started yet
            return False
        return self.next_iter > datetime.datetime.utcnow()

    @property
    def until_next(self) -> float:
        """
        Positive float with the seconds until the next iteration, based off the last
        iteration and the interval.

        If the expected time of the next iteration is in the past, this will return `0.0`
        """
        if self.next_iter is None:  # not started yet
            return 0.0

        raw_until_next = (self.next_iter - datetime.datetime.utcnow()).total_seconds()
        if raw_until_next > self.expected_interval.total_seconds():  # should never happen
            return self.expected_interval.total_seconds()
        elif raw_until_next > 0.0:
            return raw_until_next
        else:
            return 0.0

    async def sleep_until_next(self) -> None:
        """Sleep until the next iteration. Basically an "all-in-one" version of `until_next`."""
        await asyncio.sleep(self.until_next)

    def iter_start(self) -> None:
        """Register an iteration as starting."""
        self.iter_count += 1
        self.currently_running = True
        self.last_iter = datetime.datetime.utcnow()
        self.next_iter = datetime.datetime.utcnow() + self.expected_interval
        # this isn't accurate, it will be "corrected" when finishing is called

    def iter_finish(self) -> None:
        """Register an iteration as finished successfully."""
        self.currently_running = False
        # now this is accurate. imo its better to have something than nothing

    def iter_error(self, error: BaseException) -> None:
        """Register an iteration's exception."""
        self.currently_running = False
        self.last_exc_raw = error
        self.last_exc = "".join(
            traceback.format_exception(type(error), error, error.__traceback__)
        )

    def get_debug_embed(self) -> discord.Embed:
        """Get an embed with infomation on this loop."""
        raw_data: List[list] = [
            ["expected_interval", self.expected_interval],
            ["iter_count", self.iter_count],
            ["currently_running", self.currently_running],
            ["last_iter", self.last_iter or "Loop not started"],
            ["next_iter", self.next_iter or "Loop not started"],
        ]

        now = datetime.datetime.utcnow()
        processed_data: List[list]
        if self.next_iter and self.last_iter:
            processed_data = [
                ["Seconds until next", (self.next_iter - now).total_seconds()],
                ["Seconds since last", (now - self.last_iter).total_seconds()],
            ]
        else:
            processed_data = [[]]

        emoji = CHECK if self.integrity else CROSS
        embed = discord.Embed(title=f"{self.friendly_name}: `{emoji}`")
        embed.add_field(name="Raw data", value=box(tabulate.tabulate(raw_data)), inline=False)
        embed.add_field(
            name="Processed data",
            value=box(tabulate.tabulate(processed_data) or "Loop hasn't started yet."),
            inline=False,
        )
        exc = self.last_exc
        if len(exc) > 1024:
            exc = list(pagify(exc, page_length=1024))[0] + "\n..."
        embed.add_field(name="Exception", value=box(exc), inline=False)

        return embed
